"""Environment definitions."""

from j5.backends import Environment
from j5.backends.console.sr.v4 import (
    SRV4MotorBoardConsoleBackend,
    SRV4PowerBoardConsoleBackend,
    SRV4RuggeduinoConsoleBackend,
    SRV4ServoBoardConsoleBackend,
)
from j5.backends.hardware.sr.v4 import (
    SRV4MotorBoardHardwareBackend,
    SRV4PowerBoardHardwareBackend,
    SRV4RuggeduinoHardwareBackend,
    SRV4ServoBoardHardwareBackend,
)

__all__ = [
    "HARDWARE_ENVIRONMENT",
    "CONSOLE_ENVIRONMENT",
]

HARDWARE_ENVIRONMENT = Environment("Hardware Environment")

HARDWARE_ENVIRONMENT.register_backend(SRV4PowerBoardHardwareBackend)
HARDWARE_ENVIRONMENT.register_backend(SRV4ServoBoardHardwareBackend)
HARDWARE_ENVIRONMENT.register_backend(SRV4MotorBoardHardwareBackend)
HARDWARE_ENVIRONMENT.register_backend(SRV4RuggeduinoHardwareBackend)

CONSOLE_ENVIRONMENT = Environment("Console Environment")

CONSOLE_ENVIRONMENT.register_backend(SRV4PowerBoardConsoleBackend)
CONSOLE_ENVIRONMENT.register_backend(SRV4ServoBoardConsoleBackend)
CONSOLE_ENVIRONMENT.register_backend(SRV4MotorBoardConsoleBackend)
CONSOLE_ENVIRONMENT.register_backend(SRV4RuggeduinoConsoleBackend)



"""Integration with Astoria."""

import asyncio
import logging
from json import JSONDecodeError, loads
from pathlib import Path
from typing import Match, NamedTuple, Optional

from astoria.common.broadcast_event import StartButtonBroadcastEvent
from astoria.common.consumer import StateConsumer
from astoria.common.messages.astmetad import Metadata, MetadataManagerMessage
from astoria.common.messages.astprocd import ProcessManagerMessage
from astoria.common.mqtt.broadcast_helper import BroadcastHelper

LOGGER = logging.getLogger(__name__)

loop = asyncio.get_event_loop()


class GetMetadataResult(NamedTuple):
    """Result returned from fetching metadata from astoria."""

    metadata: Metadata
    usb_path: Path


class GetMetadataConsumer(StateConsumer):
    """Astoria consumer to fetch metadata."""

    name = "sr-robot3-metadata"

    def _setup_logging(self, verbose: bool, *, welcome_message: bool = True) -> None:
        """Use the logging from sr-robot3."""
        # Suppress INFO messages from gmqtt
        logging.getLogger("gmqtt").setLevel(logging.WARNING)

    def _init(self) -> None:
        """Initialise consumer."""
        self._metadata_message: Optional[MetadataManagerMessage] = None
        self._proc_message: Optional[ProcessManagerMessage] = None
        self._state_lock = asyncio.Lock()
        self._mqtt.subscribe("astmetad", self._handle_astmetad_message)
        self._mqtt.subscribe("astprocd", self._handle_astprocd_message)

    async def _handle_astmetad_message(
        self,
        match: Match[str],
        payload: str,
    ) -> None:
        """Handle astmetad status messages."""
        async with self._state_lock:
            try:
                message = MetadataManagerMessage(**loads(payload))
                if message.status == MetadataManagerMessage.Status.RUNNING:
                    LOGGER.debug("Received metadata")
                    self._metadata_message = message
                else:
                    LOGGER.warn("Cannot get metadata, astmetad is not running")
            except JSONDecodeError:
                LOGGER.error("Could not decode JSON metadata.")
            if self._metadata_message is not None and self._proc_message is not None:
                self.halt(silent=True)

    async def _handle_astprocd_message(
        self,
        match: Match[str],
        payload: str,
    ) -> None:
        """Handle astprocd status messages."""
        async with self._state_lock:
            try:
                message = ProcessManagerMessage(**loads(payload))
                if message.status == ProcessManagerMessage.Status.RUNNING:
                    LOGGER.debug("Received process info")
                    self._proc_message = message
                else:
                    LOGGER.warn("Cannot get process info, astprocd is not running")
            except JSONDecodeError:
                LOGGER.error("Could not decode JSON metadata.")
            if self._metadata_message is not None and self._proc_message is not None:
                self.halt(silent=True)

    async def main(self) -> None:
        """Main method of the command."""
        await self.wait_loop()

    @classmethod
    def get_metadata(cls) -> GetMetadataResult:
        """Get metadata."""
        gmc = cls(False, None)

        metadata = Metadata.init(gmc.config)
        path = Path("/dev/null")

        try:
            loop.run_until_complete(asyncio.wait_for(gmc.run(), timeout=0.1))
            if gmc._metadata_message is not None:
                metadata = gmc._metadata_message.metadata

            if gmc._proc_message is not None and gmc._proc_message.disk_info is not None:
                path = gmc._proc_message.disk_info.mount_path
        except ConnectionRefusedError:
            LOGGER.warning("Unable to connect to MQTT broker")
        except asyncio.TimeoutError:
            LOGGER.warning("Astoria took too long to respond, giving up.")

        return GetMetadataResult(metadata, path)


class WaitForStartButtonBroadcastConsumer(StateConsumer):
    """Wait for a start button broadcast."""

    name = "sr-robot3-wait-start"

    def _setup_logging(self, verbose: bool, *, welcome_message: bool = True) -> None:
        """Use the logging from sr-robot3."""
        # Suppress INFO messages from gmqtt
        logging.getLogger("gmqtt").setLevel(logging.WARNING)

    def _init(self) -> None:
        """
        Initialisation of the data component.

        Called in the constructor of the parent class.
        """
        self._log_event = BroadcastHelper.get_helper(
            self._mqtt,
            StartButtonBroadcastEvent,
        )

    async def main(self) -> None:
        """Send a trigger event."""
        await self._log_event.wait_broadcast()



"""Logger Setup."""

import logging
import sys


def logger_setup() -> None:
    """Setup the logger."""
    root = logging.getLogger()
    root.setLevel(logging.INFO)

    handler = logging.StreamHandler(sys.stdout)

    formatter = logging.Formatter('%(name)s %(levelname)s - %(message)s')
    handler.setFormatter(formatter)

    root.addHandler(handler)



"""sr.robot3 - Student Robotics API for Python 3."""

from astoria.common.messages.astmetad import RobotMode
from j5.boards.sr.v4 import PowerOutputPosition
from j5.components.gpio_pin import GPIOPinMode
from j5.components.motor import MotorSpecialState
from j5.components.piezo import Note

from .logging import logger_setup
from .robot import Robot, __version__

logger_setup()

COAST = MotorSpecialState.COAST
BRAKE = MotorSpecialState.BRAKE

ANALOGUE_INPUT = GPIOPinMode.ANALOGUE_INPUT
ANALOGUE_OUTPUT = GPIOPinMode.ANALOGUE_OUTPUT
INPUT = GPIOPinMode.DIGITAL_INPUT
INPUT_PULLUP = GPIOPinMode.DIGITAL_INPUT_PULLUP
INPUT_PULLDOWN = GPIOPinMode.DIGITAL_INPUT_PULLDOWN
OUTPUT = GPIOPinMode.DIGITAL_OUTPUT
PWM_OUTPUT = GPIOPinMode.PWM_OUTPUT

COMP = RobotMode.COMP
DEV = RobotMode.DEV

OUT_H0 = PowerOutputPosition.H0
OUT_H1 = PowerOutputPosition.H1
OUT_L0 = PowerOutputPosition.L0
OUT_L1 = PowerOutputPosition.L1
OUT_L2 = PowerOutputPosition.L2
OUT_L3 = PowerOutputPosition.L3

__all__ = [
    "ANALOGUE_INPUT",
    "ANALOGUE_OUTPUT",
    "BRAKE",
    "COAST",
    "COMP",
    "DEV",
    "INPUT",
    "INPUT_PULLDOWN",
    "INPUT_PULLUP",
    "Note",
    "OUTPUT",
    "OUT_H0",
    "OUT_H1",
    "OUT_L0",
    "OUT_L1",
    "OUT_L2",
    "OUT_L3",
    "PWM_OUTPUT",
    "Robot",
    "__version__",
]



"""Code to kill robot after certain amount of time."""
import logging
from signal import SIGALRM, Signals, alarm, signal
from types import FrameType

LOGGER = logging.getLogger(__name__)


def timeout_handler(signal_type: Signals, stack_frame: FrameType) -> None:
    """Handle the `SIGALRM` to kill the current process."""
    raise SystemExit("Timeout expired: Game Over!")


def kill_after_delay(timeout_seconds: int) -> None:
    """Interrupts main process after the given delay."""
    LOGGER.debug(f"Kill Signal Timeout set: {timeout_seconds}s")
    signal(SIGALRM, timeout_handler)
    alarm(timeout_seconds)



"""sr.robot3 Robot class."""

import asyncio
import logging
from datetime import timedelta
from pathlib import Path
from typing import Dict, List, Optional, Type

from astoria.common.messages.astmetad import Metadata, RobotMode
from j5 import BaseRobot, Environment
from j5 import __version__ as j5_version
from j5.backends import Backend
from j5.boards import Board, BoardGroup
from j5.boards.sr.v4 import MotorBoard, PowerBoard, ServoBoard
from j5.boards.sr.v4.ruggeduino import Ruggeduino
from j5.components.piezo import Note
from serial.tools.list_ports_common import ListPortInfo

from .astoria import GetMetadataConsumer, WaitForStartButtonBroadcastConsumer
from .env import HARDWARE_ENVIRONMENT
from .timeout import kill_after_delay

__version__ = "2022.0.0a1"

LOGGER = logging.getLogger(__name__)

loop = asyncio.get_event_loop()


class Robot(BaseRobot):
    """
    Student Robotics robot.

    Inherits from j5 BaseRobot to ensure that all boards are made safe
    in the event of a crash, and to ensure that there is only one instance
    of this class on a given machine.
    """

    def __init__(
            self,
            *,
            auto_start: bool = True,
            verbose: bool = False,
            env: Environment = HARDWARE_ENVIRONMENT,
            ignored_ruggeduinos: Optional[List[str]] = None,
    ) -> None:
        """
        Initialise a Robot object.

        :param auto_start: Automatically start the robot ignoring the start
            button, defaults to False.
        :param verbose: Turn on verbose logging, defaults to False.
        :param env: The :class:`j5.backends.environment.Environment` to run the
            Robot under. See :ref:`Environments` for more information.
        :param ignored_ruggeduinos: List of Ruggeduino serial numbers to ignore.
            See :ref:`Custom Ruggeduino Firmware` for more information.
        """
        self._auto_start = auto_start
        self._verbose = verbose
        self._environment = env

        if verbose:
            LOGGER.setLevel(logging.DEBUG)

        if ignored_ruggeduinos is None:
            self._ignored_ruggeduino_serials = []
        else:
            self._ignored_ruggeduino_serials = ignored_ruggeduinos

        LOGGER.info(f"sr.robot3 version {__version__}")
        LOGGER.debug("Verbose mode enabled.")
        LOGGER.debug(f"j5 version {j5_version}")
        LOGGER.debug(f"Environment: {self._environment.name}")

        self._init_metadata()

        self._init_power_board()
        self._init_auxilliary_boards()

        self._init_ruggeduinos()

        self._log_discovered_boards()

        if auto_start:
            LOGGER.debug("Auto start is enabled.")
            self.wait_start()
        else:
            LOGGER.debug("Auto start is disabled.")

    def _init_power_board(self) -> None:
        """
        Find and initialise the power board.

        The power board is is the only required board.
        """
        self._power_boards = self._environment.get_board_group(PowerBoard)
        self._power_board: PowerBoard = self._power_boards.singular()

        # Power on robot, so that we can find other boards.
        self.power_board.outputs.power_on()

    def _init_auxilliary_boards(self) -> None:
        """Find and initialise auxilliary boards."""
        self._motor_boards = self._environment.get_board_group(MotorBoard)
        self._servo_boards = self._environment.get_board_group(ServoBoard)

    def _init_ruggeduinos(self) -> None:
        """
        Initialise the ruggeduinos.

        Ignore any that the user has specified.
        """
        self.ignored_ruggeduinos: Dict[str, str] = {}

        ruggeduino_backend: Type[Backend] = self._environment.get_backend(Ruggeduino)

        class IgnoredRuggeduinoBackend(ruggeduino_backend):  # type: ignore
            """A backend that ignores some ruggeduinos."""

            @classmethod
            def is_arduino(cls, port: ListPortInfo) -> bool:
                """Check if a ListPortInfo represents a valid Arduino derivative."""
                if port.serial_number in self._ignored_ruggeduino_serials:
                    self.ignored_ruggeduinos[port.serial_number] = port.device
                    return False
                return super().is_arduino(port)  # type: ignore

        self._ruggeduinos = BoardGroup.get_board_group(
            Ruggeduino,
            IgnoredRuggeduinoBackend,
        )

    def _init_metadata(self) -> None:
        """Fetch metadata from Astoria."""
        self._metadata, self._code_path = GetMetadataConsumer.get_metadata()

    def _log_discovered_boards(self) -> None:
        """Log all boards that we have discovered."""
        for board in Board.BOARDS:
            LOGGER.info(f"Found {board.name} - {board.serial_number}")
            LOGGER.debug(
                f"Firmware Version of {board.serial_number}: {board.firmware_version}",
            )

    @property
    def motor_boards(self) -> BoardGroup[MotorBoard, Backend]:
        """
        Get the group of motor boards.

        :returns: a group of :class:`j5.boards.sr.v4.MotorBoard`.
        """
        return self._motor_boards

    @property
    def motor_board(self) -> MotorBoard:
        """
        Get the motor board.

        :returns: a :class:`j5.boards.sr.v4.MotorBoard`.
        :raises j5.backends.CommunicationError: there isn't exactly one
            motor board attached.
        """
        return self._motor_boards.singular()

    @property
    def ruggeduinos(self) -> BoardGroup[Ruggeduino, Backend]:
        """
        Get the group of ruggeduinos.

        :returns: a group of :class:`j5.boards.sr.v4.Ruggeduino`.
        """
        return self._ruggeduinos  # type: ignore

    @property
    def ruggeduino(self) -> Ruggeduino:
        """
        Get the ruggeduino.

        :returns: a :class:`j5.boards.sr.v4.Ruggeduino`.
        :raises j5.backends.CommunicationError: there isn't exactly one
            ruggeduino attached.
        """
        return self._ruggeduinos.singular()

    @property
    def power_board(self) -> PowerBoard:
        """
        Get the power board.

        :returns: a :class:`j5.boards.sr.v4.PowerBoard`.
        :raises j5.backends.CommunicationError: there isn't exactly one
            power board attached.
        """

    @property
    def servo_boards(self) -> BoardGroup[ServoBoard, Backend]:
        """
        Get the group of servo boards.

        :returns: a group of :class:`j5.boards.sr.v4.ServoBoard`.
        """
        return self._servo_boards

    @property
    def servo_board(self) -> ServoBoard:
        """
        Get the servo board.

        :returns: a :class:`j5.boards.sr.v4.ServoBoard`.
        :raises j5.backends.CommunicationError: there isn't exactly
            one servo board attached.
        """
        return self._servo_boards.singular()

    @property
    def metadata(self) -> Metadata:
        """
        Get the metadata information object.

        :returns: All available metadata.
        """
        return self._metadata

    @property
    def arena(self) -> str:
        """
        Determine the arena of the robot.

        If the robot is not in a match, this will be ``A``.

        :returns: The arena at the robot is in.
        """
        return self.metadata.arena

    @property
    def mode(self) -> RobotMode:
        """
        Determine the mode of the robot.

        See :ref:`Robot Modes` for available modes.

        :returns: current mode of the robot.
        """
        return self.metadata.mode

    @property
    def usbkey(self) -> Optional[Path]:
        """
        The path of the USB code drive.

        :returns: path to the mountpoint of the USB code drive.
        """
        return self._code_path

    @property
    def zone(self) -> int:
        """
        The arena zone that the robot starts in.

        :returns: arena zone that the robot starts in.
        """
        return self.metadata.zone

    def wait_start(self) -> None:
        """
        Wait for a start signal.

        Intended for use with `Robot(auto_start=False)`, to allow
        students to run code and setup their robot before the start
        of a match.
        """
        LOGGER.info("Waiting for start signal")

        astoria_start = WaitForStartButtonBroadcastConsumer(self._verbose, None)
        flash_loop = True

        async def wait_for_physical_start() -> None:
            self.power_board.piezo.buzz(timedelta(seconds=0.1), Note.A6)
            counter = 0
            led_state = False
            while not self.power_board.start_button.is_pressed and flash_loop:
                if counter % 6 == 0:
                    led_state = not led_state
                    self.power_board._run_led.state = led_state
                await asyncio.sleep(0.05)
                counter += 1
            # Turn on the LED now that we are starting
            self.power_board._run_led.state = True

        loop.run_until_complete(
            asyncio.wait(
                [
                    astoria_start.run(),
                    wait_for_physical_start(),
                ],
                return_when=asyncio.FIRST_COMPLETED,
            ),
        )

        flash_loop = False  # Stop the flashing loop

        # Reload metadata as a metadata USB may have been inserted.
        # This ensures that the game timeout is observed even if the metadata
        # USB is inserted after usercode execution begins.
        self._init_metadata()

        LOGGER.info("Start signal received; continuing.")

        if self._metadata.game_timeout is not None:
            LOGGER.info(f"Game length set to {self._metadata.game_timeout}s")
            kill_after_delay(self._metadata.game_timeout)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sr', 'sr.robot3']

package_data = \
{'': ['*']}

install_requires = \
['astoria>=0.4.0,<0.5.0', 'j5>=0.13.0,<0.14.0', 'pyserial>=3.4,<4.0']

setup_kwargs = {
    'name': 'sr.robot3',
    'version': '2022.0.0a1',
    'description': 'Student Robotics API for Python 3',
    'long_description': '# sr.robot3\n\n![Tests](https://github.com/srobo/sr-robot3/workflows/Tests/badge.svg)\n[![PyPI version](https://badge.fury.io/py/sr.robot3.svg)](https://pypi.org/project/sr.robot3/)\n[![MIT license](https://img.shields.io/badge/license-MIT-brightgreen.svg?style=flat)](https://opensource.org/licenses/MIT)\n![Bees](https://img.shields.io/badge/bees-110%25-yellow.svg)\n\nStudent Robotics API for Python 3\n\nThis package uses [PEP420]()-style namespace packaging.\n\n## Development\n\n### Requirements\n\nThis project uses the [Poetry](https://python-poetry.org) dependency and virtualenv manager.\n\nYou will also need:\n\n- Python 3.6+\n- Make\n\n### Setup\n\n- Clone the repository from GitHub to a folder on your local machine\n- `cd` to that folder, and tell Poetry to install dependencies and set up a virtualenv `poetry install`\n- You can now enter the virtual environment using `poetry shell` and develop using your IDE of choice.\n\n### Tests\n\nThe full type, test and lint suite can be run using make: `make`.\n\nYou can also run parts of the suite.\n\n- Unit tests: `make test`\n- Unit tests with HTML coverage: `make test-cov`\n- Linting: `make lint`\n- Static type checks: `make type`\n\n## Contributions\n\nThis project is released under the MIT Licence. For more information, please see LICENSE.\n\nThe CONTRIBUTORS file can be generated by executing CONTRIBUTORS.gen. This generated file contains a list of people who have contributed to sr.robot3.\n',
    'author': 'Student Robotics',
    'author_email': 'kit-brain@studentrobotics.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/srobo/sr-robot3',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
